//
//  SCAVPlayerController.swift
//  Smartclipios-app-demo
//
//  Copyright © 2018 smartclip. All rights reserved.
//

import Foundation
import AVKit
import SmartclipSDKiOS

///For the instream use case you are responsible to implement a player controller, that handles all (AV-)player related functions.
///This gives you all the control you need, to use each media player you want. This example uses an AVPlayer
class SCAVPlayerController: NSObject {

    let avPlayer                         = AVPlayer()

    private weak var sessionController   : SCAdSessionController!

    private var currentContentSource     : URL?
    private var contentPlayerItem        : AVPlayerItem?
    private var currentAdvertisementItem : AVPlayerItem?
    private var timeobserver             : Any?

    fileprivate var observedKeypathes    : Array<String> = Array()

    weak var avPlayerListener            : SCAdListener?
    weak var eventListener               : SCPlayerEventListener?

    init(with sessionController: SCAdSessionController!) {
        super.init()
        self.sessionController = sessionController
    }

    deinit {
        cleanup()
        assert(observedKeypathes.count == 0, String.init(format: "ObservedKeypathes is not empty on dealloc!: %@", observedKeypathes))
    }

    func cleanup() {
        removeAVPlayer()
        eventListener = nil
        avPlayerListener = nil
    }

    private func removeAVPlayer() {
        self.removeTimeObserver()
        self.removeContentItemObserver()
        self.removeAdItemObservers()
        self.removePlayerObserver()

        if avPlayer.timeControlStatus == .playing {
            avPlayer.pause()
            avPlayer.replaceCurrentItem(with: nil)
        }

        contentPlayerItem = nil
        currentAdvertisementItem = nil
    }

    func playContentVideo(with url: URL!) {
        currentContentSource = url
        contentPlayerItem = AVPlayerItem.init(url: url)

        self.removeAdItemObservers()
        self.addContentItemObserver()

        NotificationCenter.default.addObserver(self, selector: #selector(contentVideoFinished), name: NSNotification.Name.AVPlayerItemDidPlayToEndTime, object: self.contentPlayerItem)

        DispatchQueue.main.async { [weak self] in
            self?.avPlayer.replaceCurrentItem(with: self?.contentPlayerItem)
        }
    }

    fileprivate func addTimeObserver() {
        let updateInterval = CMTimeMake(value:1, timescale:2)

        DispatchQueue.main.async { [weak self] in
            self?.timeobserver = self?.avPlayer.addPeriodicTimeObserver(forInterval: updateInterval,
                                                                        queue: DispatchQueue.main,
                                                                        using: { time in
                                                                            self?.eventListener?.playerEventCallback(with: EventTypeTimeupdate)
            })
        }
    }

    fileprivate func removeTimeObserver() {
        guard let timeObserver = self.timeobserver else { return }

        self.avPlayer.removeTimeObserver(timeObserver)
        self.timeobserver = nil
    }

    @objc func contentVideoFinished() {
        NotificationCenter.default.removeObserver(self, name: NSNotification.Name.AVPlayerItemDidPlayToEndTime, object: self.contentPlayerItem)
        self.removeContentItemObserver()
        avPlayerListener?.contentVideoFinished?()
    }

    @objc func adItemFinished() {
        NotificationCenter.default.removeObserver(self, name: NSNotification.Name.AVPlayerItemDidPlayToEndTime, object: self.currentAdvertisementItem)
        self.removeTimeObserver()
        self.removeAdItemObservers()

        eventListener?.playerEventCallback(with: EventTypeEnded)
    }

    func stopAdSlot() {
        if avPlayer.timeControlStatus == .playing && avPlayer.currentItem == currentAdvertisementItem {
            NotificationCenter.default.removeObserver(self, name: NSNotification.Name.AVPlayerItemDidPlayToEndTime, object: self.currentAdvertisementItem)
            sessionController.stopAdSlot()
            removeTimeObserver()
            removeAdItemObservers()
            removeContentItemObserver()
            removePlayerObserver()

            contentPlayerItem = nil
            currentAdvertisementItem = nil
        }
    }

    func skipAd() {
        if avPlayer.timeControlStatus == .playing && avPlayer.currentItem == currentAdvertisementItem {
            removeAdItemObservers()
            removeTimeObserver()
            NotificationCenter.default.removeObserver(self, name: NSNotification.Name.AVPlayerItemDidPlayToEndTime, object: self.currentAdvertisementItem)
        }
    }
}

/// MARK: KVO
extension SCAVPlayerController {
    fileprivate func addPlayerObserver() {
        if !observedKeypathes.contains("playerObserver") {
            DispatchQueue.main.async {  [weak self] in
                guard let strongSelf = self else { return }
                strongSelf.avPlayer.addObserver(strongSelf, forKeyPath: "timeControlStatus", options: [.new,.old], context: nil)
                strongSelf.observedKeypathes.append("playerObserver")
            }
        }
    }

    fileprivate func removePlayerObserver() {
        if observedKeypathes.contains("playerObserver") {
            self.avPlayer.removeObserver(self, forKeyPath: "timeControlStatus")
            self.observedKeypathes.removeAll(where: { $0 == "playerObserver" })
        }
    }

    fileprivate func addContentItemObserver() {
        if !observedKeypathes.contains("contentItemObserver") {
            contentPlayerItem?.addObserver(self, forKeyPath: "status", options: [.new, .old], context: nil)
            observedKeypathes.append("contentItemObserver")
        }
    }

    fileprivate func removeContentItemObserver() {
        if observedKeypathes.contains("contentItemObserver") {
            contentPlayerItem?.removeObserver(self, forKeyPath: "status")
            observedKeypathes.removeAll(where: { $0 == "contentItemObserver" })
        }
    }

    fileprivate func addAdItemObservers() {
        addAdItemStatusObserver()
        addBufferEmptyObserver()
        addLikelyToKeepUpObserver()
    }

    fileprivate func removeAdItemObservers() {
        removeAdItemStatusObserver()
        removeBufferEmptyObserver()
        removeLikelyToKeepUpObserver()
    }

    private func addAdItemStatusObserver() {
        if !observedKeypathes.contains("adItemObserver") {
            currentAdvertisementItem?.addObserver(self, forKeyPath: "status", options: [.new, .old], context: nil)
            observedKeypathes.append("adItemObserver")
        }
    }

    private func removeAdItemStatusObserver() {
        if observedKeypathes.contains("adItemObserver") {
            currentAdvertisementItem?.removeObserver(self, forKeyPath: "status")
            observedKeypathes.removeAll(where: { $0 == "adItemObserver" })
        }
    }

    private func addBufferEmptyObserver() {
        if !observedKeypathes.contains("bufferEmptyObserver") {
            currentAdvertisementItem?.addObserver(self, forKeyPath: "playbackBufferEmpty", options: [.new], context: nil)
            observedKeypathes.append("bufferEmptyObserver")
        }
    }

    private func removeBufferEmptyObserver() {
        if observedKeypathes.contains("bufferEmptyObserver") {
            currentAdvertisementItem?.removeObserver(self, forKeyPath: "playbackBufferEmpty")
            observedKeypathes.removeAll(where: { $0 == "bufferEmptyObserver" })
        }
    }

    private func addLikelyToKeepUpObserver() {
        if !observedKeypathes.contains("likelyToKeepUpObserver") {
            currentAdvertisementItem?.addObserver(self, forKeyPath: "playbackLikelyToKeepUp", options: [.new], context: nil)
            observedKeypathes.append("likelyToKeepUpObserver")
        }
    }

    private func removeLikelyToKeepUpObserver() {
        if observedKeypathes.contains("likelyToKeepUpObserver") {
            currentAdvertisementItem?.removeObserver(self, forKeyPath: "playbackLikelyToKeepUp")
            observedKeypathes.removeAll(where: { $0 == "likelyToKeepUpObserver" })
        }
    }

    private func addPresentationSizeObserver() {
        if !observedKeypathes.contains("presentationSizeObserver") {
            currentAdvertisementItem?.addObserver(self, forKeyPath: "presentationSize", options: [.new], context: nil)
            observedKeypathes.append("presentationSizeObserver")
        }
    }

    private func removePresentationSizeObserver() {
        if observedKeypathes.contains("presentationSizeObserver") {
            currentAdvertisementItem?.removeObserver(self, forKeyPath: "presentationSize")
            observedKeypathes.removeAll(where: { $0 == "presentationSizeObserver" })
        }
    }

    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        guard let changeDict = change else { return }

        switch keyPath {
        case "timeControlStatus":
            observeAVPlayer(change: changeDict)
        case "status":
            observePlayerItem(change: changeDict)
        case "playbackBufferEmpty":
            observeBufferEmpty(change: changeDict)
        case "playbackLikelyToKeepUp":
            observeLikelyToKeepUp(change: changeDict)
        case "presentationSize":
            observePresentationSize(change: changeDict)
        default:
            break
        }
    }

    private func observeAVPlayer(change: [NSKeyValueChangeKey : Any] ) {
        guard let newState = change[NSKeyValueChangeKey.newKey] as? Int else { return }

        switch newState {
        case AVPlayer.TimeControlStatus.waitingToPlayAtSpecifiedRate.rawValue:
            break
        default: //.playing
            eventListener?.playerEventCallback(with: EventTypePlay)

            //Initialize starting of progressBar
            if avPlayer.currentItem == currentAdvertisementItem {
                if let avURLAsset = avPlayer.currentItem?.asset as? AVURLAsset {
                    avPlayerListener?.videoStartsPlayback?(with: avURLAsset.url)
                }
            }
        }
    }
    
    private func observePlayerItem(change: [NSKeyValueChangeKey : Any]) {
        guard let oldState = change[NSKeyValueChangeKey.oldKey] as? Int,
            let newState = change[NSKeyValueChangeKey.newKey] as? Int else { return }

        switch newState {
        case AVPlayerItem.Status.readyToPlay.rawValue:
            if oldState == AVPlayerItem.Status.unknown.rawValue {
                eventListener?.playerEventCallback(with: EventTypeLoadedData)
                if avPlayer.currentItem == currentAdvertisementItem {
                    eventListener?.loadAdSuccess()
                    addTimeObserver()
                }
                startPlayback()
            }
        case AVPlayerItem.Status.failed.rawValue:
            if oldState != AVPlayerItem.Status.failed.rawValue {
                if avPlayer.currentItem == currentAdvertisementItem {
                    eventListener?.loadAdFailure()
                    eventListener?.playerEventCallback(with: EventTypeError)
                } else {
                    avPlayerListener?.contentVideoError(avPlayer.currentItem?.error)
                }
            }
        default:
            break
        }
    }

    private func observeBufferEmpty(change: [NSKeyValueChangeKey : Any]) {
        guard let item = avPlayer.currentItem,
            item == currentAdvertisementItem else { return }

        if item.isPlaybackBufferEmpty {
            eventListener?.onBufferEmpty()
        }
    }

    private func observeLikelyToKeepUp(change: [NSKeyValueChangeKey : Any]) {
        guard let item = avPlayer.currentItem,
            item == currentAdvertisementItem else { return }

        if item.isPlaybackLikelyToKeepUp {
            eventListener?.onBufferRefilled()
        }
    }

    private func observePresentationSize(change: [NSKeyValueChangeKey : Any]) {
        guard let item = avPlayer.currentItem,
        item == currentAdvertisementItem else { return }

        avPlayerListener?.presentationSizeChanged(to: item.presentationSize)
    }
}

///Your playerController needs to implement the SCAdFacadeDelegate protocol
extension SCAVPlayerController: SCAdFacadeDelegate {
    func displayClickThroughAlert(_ alert: UIAlertController!) {
        avPlayerListener?.displayClickThroughAlert(alert)
    }

    func loadAd(withUrlString urlString: String?, disableSeeking: Bool) -> Bool {
        guard let urlStr = urlString,
            let url = URL.init(string: urlStr) else { return false }

        if self.currentAdvertisementItem != nil {
            NotificationCenter.default.removeObserver(self, name: NSNotification.Name.AVPlayerItemDidPlayToEndTime, object: self.currentAdvertisementItem)
            self.removeAdItemObservers()
        } else if self.contentPlayerItem != nil {
            NotificationCenter.default.removeObserver(self, name: NSNotification.Name.AVPlayerItemDidPlayToEndTime, object: self.contentPlayerItem)
        }

        self.currentAdvertisementItem = AVPlayerItem.init(url: url)
        NotificationCenter.default.addObserver(self, selector: #selector(self.adItemFinished), name: NSNotification.Name.AVPlayerItemDidPlayToEndTime, object: self.currentAdvertisementItem)
        self.addPlayerObserver()
        self.removeContentItemObserver()
        self.addAdItemObservers()

        DispatchQueue.main.async { [weak self] in
            self?.avPlayer.replaceCurrentItem(with: self?.currentAdvertisementItem)
        }

        return true
    }

    func startPlayback() {
        guard avPlayer.currentItem != nil else { return }

        if avPlayer.timeControlStatus != .playing {
            avPlayer.play()
            eventListener?.playerEventCallback(with: EventTypePlay)
        }
    }

    func pausePlayback() {
        guard avPlayer.currentItem != nil else { return }

        if avPlayer.timeControlStatus == .playing {
            avPlayer.pause()
            eventListener?.playerEventCallback(with: EventTypePause)
        }
    }

    func isPaused() -> Bool {
        return avPlayer.timeControlStatus != .playing
    }

    func mutePlayer() {
        avPlayer.isMuted = true
        eventListener?.playerEventCallback(with: EventTypeVolumeChanged)
    }

    func unmutePlayer() {
        avPlayer.isMuted = false
        eventListener?.playerEventCallback(with: EventTypeVolumeChanged)
    }

    func isMuted() -> Bool {
        return avPlayer.isMuted
    }

    func seek(_ offset: Int) {
        if (avPlayer.currentItem) != nil {
            if avPlayer.status == .readyToPlay {
                avPlayer.seek(to: CMTimeMake(value: Int64(offset), timescale: 1))
            }
        }
    }

    func getCurrentTime() -> Double {
        if avPlayer.status == .readyToPlay {
            let time = avPlayer.currentTime()
            if CMTIME_IS_INDEFINITE(time) {
                return 0
            }

            return CMTimeGetSeconds(time)
        }
        return 0
    }

    func getDuration() -> Double {
        if avPlayer.status == .readyToPlay {
            if let duration = avPlayer.currentItem?.duration {
                if CMTIME_IS_INDEFINITE(duration) {
                    return 0
                }

                return CMTimeGetSeconds(duration)
            }
        }
        return 0
    }

    func getVideoElement() -> String? {
        return nil
    }

    func setVolume(_ volume: CGFloat) {
        if avPlayer.status == .readyToPlay {
            avPlayer.volume = Float(volume)
            eventListener?.playerEventCallback(with: EventTypeVolumeChanged)
        }
    }

    func getVolume() -> CGFloat {
        var volume: CGFloat = 0.0
        if !avPlayer.isMuted {
            volume = CGFloat(avPlayer.volume)
        }

        return volume
    }

    func setCurrentContentSource() {
    }

    func getCurrentContentSource() -> String? {
        return currentContentSource?.absoluteString
    }

    func resetCurrentContentSource() -> Bool {
        if let contentItem = self.contentPlayerItem {
            self.removeTimeObserver()
            self.removeAdItemObservers()
            self.removePlayerObserver()

            NotificationCenter.default.removeObserver(self, name: NSNotification.Name.AVPlayerItemDidPlayToEndTime, object: self.currentAdvertisementItem)
            NotificationCenter.default.addObserver(self, selector: #selector(contentVideoFinished), name: NSNotification.Name.AVPlayerItemDidPlayToEndTime, object: self.contentPlayerItem)

            DispatchQueue.main.async { [weak self] in
                self?.avPlayer.replaceCurrentItem(with: contentItem)
                self?.addContentItemObserver()
                self?.startPlayback()
            }
        }

        return true
    }

    func getAdInfo(_ completion: SCAdInfoCompletionBlock!) {
        eventListener?.getAdInfo { (adInfo) in
            completion(adInfo)
        }
    }


    func getAdError(_ completion: SCAdErrorCompletionBlock!) {
        eventListener?.getAdError { (adError) in
            completion(adError)
        }
    }

    func getPublicAdSlot(_ completion: SCPublicAdSlotCompletionBlock!) {
        eventListener?.getPublicAdSlot { (publicAdSlot) in
            completion(publicAdSlot)
        }
    }

    func onClickThrough() {
        eventListener?.onClickThrough()
    }

    func onEventCallback(_ event: SCAdEvent?) {
        avPlayerListener?.onEventCallback(with: event)
    }

    func getError() -> [AnyHashable : Any]? {
        if avPlayer.status == AVPlayer.Status.failed {
            if let error = avPlayer.error {
                let mediaError = SCMediaError.init(errorCode: error._code, message: error.localizedDescription)
                return mediaError.description()
            }
        } else if avPlayer.currentItem?.status == AVPlayerItem.Status.failed {
            if let error = avPlayer.currentItem?.error {
                let mediaError = SCMediaError.init(errorCode: error._code, message: error.localizedDescription)
                return mediaError.description()
            }
        }

        return nil
    }

    func onVisibilityChanged(_ visible: Bool) {
        print("onVisibilityChanged")
        avPlayerListener?.visibilityChanged(visible)
    }
}
