//
//  SCSequencerViewController.swift
//  Smartclipios-app-demo
//
//  Copyright © 2019 smartclip. All rights reserved.
//

import UIKit
import Foundation
import AVKit
import SmartclipSDKiOS

/// Use this as an example for a view controller for instream playback
/// It uses a SCAdSequencer object for playback control and adSlot timing
class SCSequencerViewController: UIViewController {
    private let playerViewController = AVPlayerViewController()
    private let sessionController    = SCAdSessionController()
    private var avPlayerController   : SCAVPlayerController?
    private let progressBar          = UIProgressView()
    private var skipButton           = UIButton(type: .custom)

    private var sequencer           : SCAdSequencer?
    private var bumper              = Constants.bumper
    private var opener              = Constants.opener
    private var closer              = Constants.closer
    private var contentUrl          = Constants.contentVideoURL
    private var preRollUrl          = Constants.adURL
    private var midRollUrl          = Constants.adURL2
    private var postRollUrl         = Constants.adURL
    private var titleLabel          = UILabel()
    private var progressTimer       : Timer?
    private var clickListener       : UITapGestureRecognizer?
    private var clickThroughView    : UIView?

    private var allOverDuration         = -1.0
    private var recentClipsTime         = 0.0

    deinit {
        /// Cleanup on deinit!
        sequencer?.cleanup()
        sessionController.cleanup()
    }

    override func viewDidLoad() {
        super.viewDidLoad()
        sessionController.sessionStateListener = self as SCAdSessionStateListener

        /// The SCAVPlayerController is just an example implementation of a controller
        /// that uses the AVPlayer for video playback, you could as well create a different implementation
        /// but your player controller has to implement the SCAdFacadeDelegate protocol
        avPlayerController = SCAVPlayerController.init(with: sessionController)
        setupSequencer()
        setupAVPlayerViewController()
    }

    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        sequencer?.pause()
    }

    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        guard sequencer?.isPaused() ?? false else { return }

        sequencer?.resumeSequencer()
    }

    /// View positioning
    override func viewDidLayoutSubviews() {
        super.viewDidLayoutSubviews()

        titleLabel.sizeToFit()

        let videoViewFrame = playerViewController.view.frame
        skipButton.frame = CGRect.init(x: view.frame.size.width - 100, y: videoViewFrame.origin.y + videoViewFrame.size.height - 50, width: 90, height: 30)

        clickThroughView?.frame = videoViewFrame
    }

    /// Orientation changes
    override func viewWillTransition(to size: CGSize, with coordinator: UIViewControllerTransitionCoordinator) {
        coordinator.animate(alongsideTransition: { [weak self] (UIViewControllerTransitionCoordinatorContext) -> Void in
            self?.view.setNeedsLayout()
        }, completion: { [weak self] (UIViewControllerTransitionCoordinatorContext) -> Void in
            self?.view.setNeedsLayout()
        })

        super.viewWillTransition(to: size, with: coordinator)
    }

    /// tell the sequencer about the content video
    /// give it the SCAdConfiguration and a list of SCAdSlot models
    private func setupSequencer() {
        /// Create the sequnecer and give it access to the playerController through its facade protocol implementation
        if let facadeDelegate = avPlayerController as SCAdFacadeDelegate? {
            sequencer = SCAdSequencer.init(playerDelegate: facadeDelegate, sessionController: sessionController)

            /// Set this viewController as the sequencers delegate
            sequencer?.delegate = self

            sequencer?.contentURL = contentUrl
            sequencer?.configuration = self.configuration()
            sequencer?.adSlots = self.adSlots()

            /// the sequencer needs to listen to some player events
            avPlayerController?.eventListener = sequencer as? SCPlayerEventListener
            avPlayerController?.avPlayerListener = sequencer as? SCAdListener
        }
    }

    /// Create and return your adapted configuration
    private func configuration() -> SCAdConfiguration {
        let macros = SCAdMacros.init()

        // we are only using the default macro settings here,
        // you should set your desired values on the macros object
        macros.appBundle = Bundle.main.bundleIdentifier!

        let environment = SCAdEnvironment.init(macros: macros)
        environment.desiredBitrate = 2000
        environment.desiredMimeTypes = ["video/mp4"]

        let config = SCAdConfiguration.defaultInstreamConfiguration(with:environment)
        config.customTitle = "Werbung"
        config.customTitleColor = UIColor.yellow
        config.customProgressBarColor = UIColor.yellow

        config.clickType = ClickableWithConfirmationDialog

/* if you want adVerification, then add your partnerName and omidVersion like this:
//
//         config.omidVersion = String(SmartclipSDKiOSVersionNumber)
//         config.omidPartnerName = "Smartclip"
//
*/

        return config
    }

    /// Create and define your adSlots
    private func adSlots() -> NSMutableArray {
        let variants = SCAdVariants.init(opener: opener, closer: closer, bumper: bumper)
        let slots = NSMutableArray.init()

        slots.add(SCAdSlot.init(adTag: preRollUrl, relativeSlotTime: 0.0, variants: variants, skipOffset:3, breakPosition: SCBreakPositionPreRoll))
        slots.add(SCAdSlot.init(adTag: midRollUrl, relativeSlotTime: 0.5, variants: variants, skipOffset:3, breakPosition: SCBreakPositionMidRoll))
        slots.add(SCAdSlot.init(adTag: postRollUrl, relativeSlotTime: 1.0, variants: variants, skipOffset:3, breakPosition: SCBreakPositionPostRoll))

        return slots
    }

    /// This view controller uses an avPlayerViewController for the avPlayer, you could use something different though
    private func setupAVPlayerViewController() {
        addChild(playerViewController)

        let playerView: UIView = playerViewController.view
        view.addSubview(playerView)

        playerViewController.didMove(toParent: self)
        playerView.translatesAutoresizingMaskIntoConstraints = false

        if #available(iOS 11.0, *) {
            NSLayoutConstraint.activate([
                playerView.centerYAnchor.constraint(equalTo: view.centerYAnchor),
                playerView.centerXAnchor.constraint(equalTo: view.centerXAnchor),
                playerView.widthAnchor.constraint(equalTo: view.safeAreaLayoutGuide.widthAnchor, multiplier: 1),
                playerView.heightAnchor.constraint(equalTo: playerView.widthAnchor, multiplier: 9/16)
            ])
        } else {
            // Fallback on earlier versions
            NSLayoutConstraint.activate([
            playerView.centerYAnchor.constraint(equalTo: view.centerYAnchor),
            playerView.centerXAnchor.constraint(equalTo: view.centerXAnchor),
            playerView.widthAnchor.constraint(equalTo: view.widthAnchor, multiplier: 1),
            playerView.heightAnchor.constraint(equalTo: playerView.widthAnchor, multiplier: 9/16)])
        }

        playerViewController.player = avPlayerController?.avPlayer

        if let overlayView = playerViewController.contentOverlayView {
            titleLabel = UILabel.init()
            titleLabel.shadowColor = .black
            titleLabel.shadowOffset = CGSize(width: 1, height: 1)
            titleLabel.isHidden = true
            titleLabel.translatesAutoresizingMaskIntoConstraints = false
            overlayView.addSubview(titleLabel)

            NSLayoutConstraint.activate([
                titleLabel.leadingAnchor.constraint(equalTo: overlayView.leadingAnchor, constant: 10),
                titleLabel.topAnchor.constraint(equalTo: overlayView.topAnchor, constant: 10 ),
            ])

            overlayView.addSubview(progressBar)
            progressBar.isHidden = true
            progressBar.tintColor = UIColor.white
            progressBar.trackTintColor = UIColor.clear
            progressBar.translatesAutoresizingMaskIntoConstraints = false
            overlayView.addSubview(progressBar)

            NSLayoutConstraint.activate([
                progressBar.leadingAnchor.constraint(equalTo: overlayView.leadingAnchor, constant: 0),
                progressBar.bottomAnchor.constraint(equalTo: overlayView.bottomAnchor, constant: -10 ),
                progressBar.widthAnchor.constraint(equalTo: overlayView.widthAnchor, multiplier: 1),
                progressBar.heightAnchor.constraint(equalToConstant: progressBar.frame.size.height)
            ])

            sequencer?.addView(forVisibilityCheck: playerView)
        }

        clickListener = UITapGestureRecognizer(target: self, action: #selector(onClick))
        clickListener?.isEnabled = false
        
        if #available(iOS 12, *) {
            playerView.addGestureRecognizer(clickListener!)
        } else {
            clickThroughView = UIView.init(frame: playerView.frame)
            view.addSubview(clickThroughView!)
            clickThroughView?.addGestureRecognizer(clickListener!)
        }

        skipButton = UIButton.init(type: .custom)
        skipButton.setTitle("Skip Ad", for: .normal)
        skipButton.setTitleColor(UIColor.white, for: .normal)
        skipButton.backgroundColor = UIColor.red
        skipButton.isHidden = true
        skipButton.addTarget(self, action: #selector(onSkipPressed(_:)), for: .touchDown)

        view.addSubview(skipButton)
        view.bringSubviewToFront(skipButton)
    }

    /// Clickthrough action of a gestureRecognizer
    @objc func onClick() {
        sequencer?.onClickThrough()
    }

    /// Skip button action
    @objc func onSkipPressed(_ sender: UIButton?) {
        avPlayerController?.getAdInfo { [weak self] (adInfo) in
            if let unwrappedAdInfo = adInfo {
                if unwrappedAdInfo.variant == "commercial" {
                    if let recentTime = self?.recentClipsTime, let duration = self?.avPlayerController?.getDuration() {
                        self?.recentClipsTime = recentTime + duration
                    }
                }
            }
        }
        deleteProgressTimer()
        sequencer?.skipAd()
        skipButton.isHidden = true
    }

    /// Helper function
    func displayAlert(tite:String?, message:String?) {
        let alertController = UIAlertController.init(title: tite, message: message, preferredStyle: .alert)
        let action = UIAlertAction.init(title: "OK", style: .cancel, handler: nil)
        alertController.addAction(action)
        self.present(alertController, animated: true, completion: nil)
    }

    //override point for finished content video
    func onContentVideoFinished() {}

    //override point for finished advertisement playback
    func adSlotFinished() {
        playerViewController.showsPlaybackControls = true
        titleLabel.isHidden = true
        progressBar.isHidden = true

        clickListener?.isEnabled = false

        progressTimer?.invalidate()
        progressTimer = nil
        skipButton.isHidden = true

        clickThroughView?.isHidden = true
        recentClipsTime = 0.0
        allOverDuration = -1.0
    }

    //override point for starting advertisement playback
    func adSlotStarted(){
        playerViewController.showsPlaybackControls = false
        if let config = sequencer?.configuration {
            titleLabel.text = config.customTitle
            titleLabel.textColor = config.customTitleColor
        }
        titleLabel.sizeToFit()

        titleLabel.isHidden = false
        skipButton.isHidden = true

        progressBar.progress = 0.0
        progressBar.isHidden = false

        clickListener?.isEnabled = true

        clickThroughView?.isHidden = false

        playerViewController.contentOverlayView?.bringSubviewToFront(titleLabel)
    }

    /// When the content video is finished, this function will be called
    func contentVideoFinished() {
        onContentVideoFinished()
    }

    /// Helper function to update progressbar during ad playback
    func createProgressTimer() {
        if progressTimer != nil { return }

        var slotDuration = self.allOverDuration
        let width = playerViewController.view.frame.size.width
        var updateInterval = 0.0

        avPlayerController?.getAdInfo { [weak self] (adInfo) in
            if let unwrappedAdInfo = adInfo {
                if unwrappedAdInfo.variant == "commercial" {
                    self?.avPlayerController?.getPublicAdSlot { [weak self] (publicAdSlot) in
                        if let unwrappedPublicAdSlot = publicAdSlot {
                            if slotDuration <= 0.0  {
                                self?.recentClipsTime = 0
                                slotDuration = Double(unwrappedPublicAdSlot.totalCommercialsDuration)
                                updateInterval = slotDuration > 0.0 ? (slotDuration / (Double)(width * UIScreen.main.scale)) : 0.0
                                self?.allOverDuration = slotDuration
                            }
                            self?.progressBar.isHidden = false

                            self?.progressTimer = Timer.scheduledTimer(withTimeInterval: updateInterval, repeats: true, block: { [weak self] (Timer) in
                                self?.updateProgress()
                            })
                        }
                    }
                } else {
                    self?.deleteProgressTimer()
                    self?.progressBar.isHidden = true
                }
            }
        }
    }

    func totalDurationCurrentTime() -> Double {
        guard let duration = avPlayerController?.getCurrentTime() else { return 0.0 }
        return recentClipsTime + duration
    }

    func deleteProgressTimer() {
        progressTimer?.invalidate()
        progressTimer = nil
    }

    func updateProgress() {
        let totalCommercialsCurrentTime = self.totalDurationCurrentTime()

        if totalCommercialsCurrentTime >= 0.0 {
            let realProgress = Double(totalCommercialsCurrentTime/self.allOverDuration)

            progressBar.setProgress(Float(realProgress), animated: true)
        }
        view.setNeedsLayout()
    }
}

extension SCSequencerViewController: SCAdSessionStateListener {
    /// Callback function that indicates that the SmartclipSDKiOS is ready to use
    func onSDKReady() {
        sequencer?.start()
    }
}

extension SCSequencerViewController: SCAdSequencerDelegate {
    ///override point for changed presentationSize (Called when the presentationSize of the AVItem changes)
    func presentationSizeChanged(to newSize: CGSize) {
    }

    /// Sequncer callback function to start the content video
    func playContentVideo(with urlString: String) {
        avPlayerController?.playContentVideo(with: URL.init(string: urlString))
    }

    /// Sequncer callback function to pause the content video
    func pauseContentVideo() {
        avPlayerController?.pausePlayback()
    }

    /// Sequncer callback function to resume the content video
    func resumeContentVideo() {
        avPlayerController?.startPlayback()
    }

    ///Slot reinsertion (example implementation - adjust to your own needs
    func userDidScrub(_ removedAdSlots: [SCAdSlot], currentRelativePosition: CGFloat) -> [SCAdSlot] {
        if currentRelativePosition <= 1.0 {
            var reinsertedSlots = Array<SCAdSlot>()
            var index = 1

            for adSlot in removedAdSlots {
                let newValue = (currentRelativePosition + 0.1 * CGFloat(index)) > 1.0 ? 1.0 : (currentRelativePosition + 0.1 * CGFloat(index))

                adSlot.relativeSlotTime = newValue
                reinsertedSlots.append(adSlot)
                index = index + 1
            }

            return reinsertedSlots
        }
        return []
    }

    /// If the playback or loading of the content video throws an error, it will be
    /// received here.
    func contentVideoError(_ error: Error) {
        displayAlert(tite: "Content Video Error!",
                     message: String(format: "An error occured while trying to display the desired content video. \nMessage: %@", error.localizedDescription));
    }

    /// Callback method for SCAdEvents
    func onEventCallback(with info:SCAdEvent) {
        switch info.type {
        case ON_AD_SLOT_STARTED:
            self.adSlotStarted()
        case ON_AD_SLOT_FINISHED:
            self.adSlotFinished()
        case ON_AD_SKIPPABLE_STATE_CHANGE:
            self.skipButton.isHidden = false
        case ON_AD_STARTED:
            self.skipButton.isHidden = true
        case ON_AD_PLAYBACK_FINISHED:
            avPlayerController?.getAdInfo { [weak self] (adInfo) in
                if let unwrappedAdInfo = adInfo {
                    if unwrappedAdInfo.variant == "commercial" {
                        if let recentTime = self?.recentClipsTime, let duration = self?.avPlayerController?.getDuration() {
                            self?.recentClipsTime = recentTime + duration
                        }
                    }
                }
            }
            self.deleteProgressTimer()
        case ON_AD_PAUSED:
            self.deleteProgressTimer()
        case ON_AD_ERROR:
            avPlayerController?.getAdError { (adError) in
                if let unwrappedAdError = adError {
                    NSLog("adError: \(unwrappedAdError.errorDescription) code: \(unwrappedAdError.errorCode)")
                }
            }
        default:
            break
        }
    }

    /// When the sequencer has finished his sequence, this function is being called
    func sequencerFinished() {
        /// Do something, when the sequencer has finished...
    }

    /// If the user clicks on the ad and clickThrough is active the sequencer
    /// tells its delegate to open it
    func displayClickThroughAlert(_ alert: UIAlertController) {
        self.present(alert, animated: true, completion: nil)
    }

    func adStartsPlayback(with url: URL) {
        if url.absoluteString != opener &&
            url.absoluteString != closer &&
            url.absoluteString != bumper &&
            url.absoluteString != contentUrl {
            createProgressTimer()
        } else {
            deleteProgressTimer()
            progressBar.isHidden = true
        }
    }
}
